"""Compatibility layer for QCElemental (QCSchema)."""

from typing import Any, Dict

from qcio import ProgramInput, SinglePointResults, Wavefunction


def to_qcel_input(prog_input: ProgramInput) -> Dict[str, Any]:
    """Return the QCElemental v1 input schema representation of the input
    (AtomicInput dict).

    Args:
        prog_input: The input object to convert.

    Returns:
        The QCElemental v1 dict representation of an AtomicInput object.
    """
    return {
        "molecule": {
            "symbols": prog_input.structure.symbols,
            "geometry": prog_input.structure.geometry,
            "molecular_charge": prog_input.structure.charge,
            "molecular_multiplicity": prog_input.structure.multiplicity,
            "connectivity": prog_input.structure.connectivity or None,
            # Insane defaults, but they are the defaults in qcelemental
            # Setting to True instead so qcel doesn't rotate or translate the molecule
            # https://github.com/MolSSI/QCElemental/blob/8e5a8cff52a6438ff9d6c1c6bbf1aeb4f02f12e1/qcelemental/models/molecule.py#L262-L281  # noqa: E501
            "fix_com": True,
            "fix_orientation": True,
            "identifiers": prog_input.structure.identifiers.model_dump(
                exclude={"name_IUPAC", "name", "extras"}
            ),  # not on qcel model
        },
        "driver": prog_input.calctype,
        "model": prog_input.model.model_dump(exclude={"extras"}),
        "keywords": prog_input.keywords,
        "extras": prog_input.extras,
    }


def from_qcel_output_results(
    qcel_output: Dict[str, Any],
) -> SinglePointResults:
    """Create a SinglePointSuccessfulOutput or SinglePointFailedOutput from the
    QCElemental v1 output schema representation of the output (AtomicResult dict).

    Args:
        qcel_output: The QCElemental v1 output schema representation of the output.
            May be a dict representing an AtomicResult or FailedOperation.
    """
    # Collect values from keys that exist in qcio
    qcio_to_qcel = {
        "calcinfo_natoms": "calcinfo_natom",
        "energy": "return_energy",
        "gradient": "return_gradient",
        "hessian": "return_hessian",
    }
    results = {}
    for key in SinglePointResults.__annotations__:
        if key in qcio_to_qcel:
            qcel_key = qcio_to_qcel[key]
        else:
            qcel_key = key

        value = qcel_output["properties"].get(qcel_key)
        if value is not None:
            results[key] = value

    # Override with return_result as qcel may not have save the key value to .properties
    results[qcel_output["driver"]] = qcel_output["return_result"]

    if qcel_output["wavefunction"]:
        results["wavefunction"] = {
            key: value
            for key, value in qcel_output["wavefunction"].items()
            if key in Wavefunction.__annotations__
        }

    results["extras"] = {"extras": {"NOTE": "Results computed using QCEngine"}}
    return SinglePointResults(**results)



# https://physics.nist.gov/cgi-bin/cuu/Value?bohrrada0
BOHR_TO_ANGSTROM = 0.529177210544
ANGSTROM_TO_BOHR = 1 / BOHR_TO_ANGSTROM
# https://www.physics.nist.gov/cgi-bin/cuu/Value?hr
HARTREE_TO_JOULE = 4.3597447222060e-18
# https://physics.nist.gov/cgi-bin/cuu/Value?na
AVOGADRO_NUMBER = 6.02214076e23
# https://www.nist.gov/pml/special-publication-811/nist-guide-si-appendix-b-conversion-factors/nist-guide-si-appendix-b8
# Using kcal(th) as Therochemical calorie is the unit used in quantum chemistry.
KCAL_TO_JOULE = 4.184e3
HARTREE_TO_KCAL_PER_MOL = HARTREE_TO_JOULE / KCAL_TO_JOULE * AVOGADRO_NUMBER

# Declaring Private so I can update implementations of this towards a periodic table.
# Public features are exposed on core models. e.g., Structure.atomic_numbers
_ELEMENTS = {
    "H": 1,
    "He": 2,
    "Li": 3,
    "Be": 4,
    "B": 5,
    "C": 6,
    "N": 7,
    "O": 8,
    "F": 9,
    "Ne": 10,
    "Na": 11,
    "Mg": 12,
    "Al": 13,
    "Si": 14,
    "P": 15,
    "S": 16,
    "Cl": 17,
    "Ar": 18,
    "K": 19,
    "Ca": 20,
    "Sc": 21,
    "Ti": 22,
    "V": 23,
    "Cr": 24,
    "Mn": 25,
    "Fe": 26,
    "Co": 27,
    "Ni": 28,
    "Cu": 29,
    "Zn": 30,
    "Ga": 31,
    "Ge": 32,
    "As": 33,
    "Se": 34,
    "Br": 35,
    "Kr": 36,
    "Rb": 37,
    "Sr": 38,
    "Y": 39,
    "Zr": 40,
    "Nb": 41,
    "Mo": 42,
    "Tc": 43,
    "Ru": 44,
    "Rh": 45,
    "Pd": 46,
    "Ag": 47,
    "Cd": 48,
    "In": 49,
    "Sn": 50,
    "Sb": 51,
    "Te": 52,
    "I": 53,
    "Xe": 54,
    "Cs": 55,
    "Ba": 56,
    "La": 57,
    "Ce": 58,
    "Pr": 59,
    "Nd": 60,
    "Pm": 61,
    "Sm": 62,
    "Eu": 63,
    "Gd": 64,
    "Tb": 65,
    "Dy": 66,
    "Ho": 67,
    "Er": 68,
    "Tm": 69,
    "Yb": 70,
    "Lu": 71,
    "Hf": 72,
    "Ta": 73,
    "W": 74,
    "Re": 75,
    "Os": 76,
    "Ir": 77,
    "Pt": 78,
    "Au": 79,
    "Hg": 80,
    "Tl": 81,
    "Pb": 82,
    "Bi": 83,
    "Po": 84,
    "At": 85,
    "Rn": 86,
    "Fr": 87,
    "Ra": 88,
    "Ac": 89,
    "Th": 90,
    "Pa": 91,
    "U": 92,
    "Np": 93,
    "Pu": 94,
    "Am": 95,
    "Cm": 96,
    "Bk": 97,
    "Cf": 98,
    "Es": 99,
    "Fm": 100,
    "Md": 101,
    "No": 102,
    "Lr": 103,
    "Rf": 104,
    "Db": 105,
    "Sg": 106,
    "Bh": 107,
    "Hs": 108,
    "Mt": 109,
    "Ds": 110,
    "Rg": 111,
    "Cn": 112,
    "Nh": 113,
    "Fl": 114,
    "Mc": 115,
    "Lv": 116,
    "Ts": 117,
    "Og": 118,
}

# https://depts.washington.edu/eooptic/linkfiles/dielectric_chart%5B1%5D.pdf
_SOLVENTS_DIELECTRIC = {
    "acetic acid": 6.15,
    "acetone": 20.7,
    "acetonitrile": 37.5,
    "anisole": 4.33,
    "benzene": 2.27,
    "bromobenzene": 5.17,
    "carbon disulfide": 2.6,
    "carbon tetrachloride": 2.24,
    "chlorobenzene": 5.62,
    "chloroform": 4.81,
    "cyclohexane": 2.02,
    "dibutyl ether": 3.1,
    "o -dichlorobenzene": 9.93,
    "1,2-dichloroethane": 10.36,
    "dichloromethane": 8.93,
    "diethylamine": 3.6,
    "diethyl ether": 4.33,
    "1,2-dimethoxyethane": 7.2,
    "n,n -dimethylacetamide": 37.8,
    "n,n -dimethylformamide": 36.7,
    "dimethyl sulfoxide": 46.7,
    "1,4-dioxane": 2.25,
    "ethanol": 24.5,
    "ethyl acetate": 6.02,
    "ethyl benzoate": 6.02,
    "formamide": 111,
    "hexamethylphosphoramide": 30,
    "isopropyl alcohol": 17.9,
    "methanol": 32.7,
    "2-methyl-2-propanol": 10.9,
    "nitrobenzene": 34.82,
    "nitromethane": 35.87,
    "pyridine": 12.4,
    "tetrahydrofuran": 7.58,
    "toluene": 2.38,
    "trichloroethylene": 3.4,
    "triethylamine": 2.42,
    "trifluoroacetic acid": 8.55,
    "2,2,2-trifluoroethanol": 8.55,
    "water": 80.1,
    "o -xylene": 2.57,
}



"""Utility functions for the models module."""

import warnings
from typing import Any, Dict

from ..constants import ANGSTROM_TO_BOHR


def deprecated_function(new_name: str):
    """Notify users that a function is deprecated and will be removed in the future.
    Works equally well on functions and methods.

    Usage:
        @deprecated("new_name")
        def old_name():
            pass
    """

    def decorator(func):
        def wrapped(*args, **kwargs):
            warnings.warn(
                f"{func.__name__} is deprecated and will be removed in future "
                f"versions. Please use {new_name} instead.",
                FutureWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapped

    return decorator


def deprecated_class(new_name: str):
    """Notify users that a class is deprecated and will be removed in the future.
    Works only on Pydantic models since it relies on the model_post_init method.

    Usage:
        @deprecated_class("new_name")
        class OldName:
            pass
    """

    def class_decorator(cls):
        orig_init = cls.__init__

        def new_init(self, *args, **kwargs):
            warnings.warn(
                f"{cls.__name__} is deprecated and will be removed in a future "
                f"release. Please use '{new_name}' instead.",
                category=FutureWarning,
                stacklevel=2,
            )
            orig_init(self, *args, **kwargs)

        cls.__init__ = new_init
        return cls

    return class_decorator


def renamed_class(new_cls: Any):
    """Notify users that a class is deprecated and has been renamed.

    Usage:
        @renamed_class(NewClass)
        class OldClass:
            pass
    """

    def decorator(cls):
        def return_new_cls(cls, *args, **kwargs):
            warnings.warn(
                f"{cls.__name__} is deprecated and and will be removed in a future "
                f"release. Please use '{new_cls.__name__}' instead.",
                category=FutureWarning,
                stacklevel=2,
            )
            return new_cls(*args, **kwargs)

        cls.__new__ = return_new_cls
        return cls

    return decorator


def smiles_to_structure(
    smiles: str, program: str = "rdkit", force_field: str = "UFF"
) -> Dict[str, Any]:
    """Convert a SMILES string to a Structure object in dictionary form.

    Args:
        smiles: The SMILES string to convert.
        method: The method to use for the conversion. Defaults to "UFF".

    Returns:
        A dictionary with the keys "symbols" and "geometry".
    """
    SUPPORTED_PROGRAMS = ["rdkit"]

    # Remove newline characters if present
    smiles = smiles.strip()

    if program not in SUPPORTED_PROGRAMS:
        raise ValueError(f"Unsupported program: '{program}'.")

    if program == "rdkit":
        try:
            from rdkit import Chem
            from rdkit.Chem import AllChem
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "The 'rdkit' module is required for this function. "
                "Please install it with 'pip install qcio[rdkit]'."
            )
        else:
            # Convert SMILES to RDKit Mol object
            mol = Chem.MolFromSmiles(smiles)
            assert mol is not None, f"Failed to convert SMILES to RDKit Mol: {smiles}"
            canonical_smiles = Chem.MolToSmiles(mol, canonical=True)
            mol = Chem.AddHs(mol)

            # Generate 3D coordinates
            AllChem.EmbedMolecule(mol, AllChem.ETKDG())
            # Optimize the molecule using the specified force field
            if force_field.upper() == "UFF":
                AllChem.UFFOptimizeMolecule(mol)
            elif force_field.upper() == "MMFF94":
                AllChem.MMFFOptimizeMolecule(mol, mmffVariant="MMFF94")  # type: ignore
            elif force_field.upper() == "MMFF94S":
                AllChem.MMFFOptimizeMolecule(mol, mmffVariant="MMFF94s")  # type: ignore
            else:
                raise ValueError(f"Unsupported force_field: {force_field}")

            # Get atom symbols
            atoms = [atom.GetSymbol() for atom in mol.GetAtoms()]  # type: ignore

            # Get atom positions
            conf = mol.GetConformer()
            geometry_angstrom = conf.GetPositions()
            geometry_bohr = geometry_angstrom * ANGSTROM_TO_BOHR

    return {
        "symbols": atoms,
        "geometry": geometry_bohr,
        "charge": Chem.GetFormalCharge(mol),
        "identifiers": {"canonical_smiles": canonical_smiles, "smiles": smiles},
    }



"""End user output and results objects from a calculation."""

from __future__ import annotations

import sys
import warnings
from itertools import product
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    get_args,
)

import numpy as np
from pydantic import field_validator, model_validator
from typing_extensions import Self

from qcio.helper_types import SerializableNDArray

from .base_models import CalcType, Files, Provenance, QCIOModelBase
from .inputs import DualProgramInput, FileInput, Inputs, InputType, ProgramInput
from .structure import Structure
from .utils import deprecated_class

if TYPE_CHECKING:  # pragma: no cover
    pass


__all__ = [
    "SinglePointResults",
    "Wavefunction",
    "OptimizationResults",
    "ProgramOutput",
    "ResultsType",
    "Results",
    "NoResults",
    "SinglePointOutput",
    "ProgramFailure",
    "OptimizationOutput",
]


class Wavefunction(QCIOModelBase):
    """The wavefunction for a single point calculation.

    Attributes:
        scf_eigenvalues_a: The SCF alpha-spin orbital eigenvalues.
        scf_eigenvalues_b: The SCF beta-spin orbital eigenvalues.
        scf_occupations_a: The SCF alpha-spin orbital occupations.
        scf_occupations_b: The SCF beta-spin orbital occupations.
    """

    scf_eigenvalues_a: Optional[SerializableNDArray] = None
    scf_eigenvalues_b: Optional[SerializableNDArray] = None
    scf_occupations_a: Optional[SerializableNDArray] = None
    scf_occupations_b: Optional[SerializableNDArray] = None

    @field_validator(
        "scf_eigenvalues_a",
        "scf_eigenvalues_b",
        "scf_occupations_a",
        "scf_occupations_b",
    )
    @classmethod
    def to_numpy(cls, val, _info) -> Optional[np.ndarray]:
        return np.asarray(val) if val is not None else None


class NoResults(QCIOModelBase):
    """Am empty results object for when no results are present.

    This is used to avoid having to use Optional[Results] in ProgramOutput which gives
    greater type safety by ensuring that results are always present. Also, this object
    functions as an empty container so that end user code doesn't have to check if
    results are None before accessing them--akin to returning an empty list instead of
    None for a function.
    """

    def __bool__(self):
        """Behave like an empty container."""
        return False


class SinglePointResults(QCIOModelBase):
    """The computed results from a single point calculation.

    Attributes:
        calcinfo_natoms: The number of atoms as computed by the program.
        calcinfo_nalpha: The number of alpha electrons as computed by the program.
        calcinfo_nbeta: The number of beta electrons as computed by the program.
        calcinfo_nbasis: The number of basis functions in the calculation.
        calcinfo_nmo: The number of molecular orbitals in the calculation

        energy: The electronic energy of the structure in Hartrees.
        gradient: The gradient of the structure in Hartrees/Bohr.
        hessian: The hessian of the structure in Hartrees/Bohr^2.
        nuclear_repulsion_energy: The nuclear repulsion energy of the structure in
            Hartrees.

        wavefunction: Wavefunction data from the calculation.

        freqs_wavenumber: The frequencies of the structure in wavenumbers.
        normal_modes_cartesian: 3D n_vibmodes x n_atoms x 3 array containing
            un-mass-weighted Cartesian displacements of each normal mode.
        gibbs_free_energy: Gibbs free energy (i.e. thermochemical analysis) in Hartrees
            of a system where translation / rotation / vibration degrees of freedom are
            approximated using ideal gas / rigid rotor / harmonic oscillator
            respectively.
        scf_dipole_moment: The x, y, z component of the dipole moment of the structure
            in units of e a0 (NOT Debye!).

    """

    # calcinfo contains general information about the calculation
    calcinfo_natoms: Optional[int] = None
    calcinfo_nbasis: Optional[int] = None
    calcinfo_nmo: Optional[int] = None
    calcinfo_nalpha: Optional[int] = None
    calcinfo_nbeta: Optional[int] = None

    # Core properties
    energy: Optional[float] = None
    gradient: Optional[SerializableNDArray] = None  # Coerced to 2D array
    hessian: Optional[SerializableNDArray] = None  # Coerced to 2D array
    nuclear_repulsion_energy: Optional[float] = None

    # Wavefunction data
    wavefunction: Optional[Wavefunction] = None

    # Frequency data
    freqs_wavenumber: List[float] = []
    normal_modes_cartesian: Optional[SerializableNDArray] = None  # Coerced to 3D array
    gibbs_free_energy: Optional[float] = None

    # SCF results
    scf_dipole_moment: Optional[List[float]] = None

    @field_validator("normal_modes_cartesian")
    @classmethod
    def validate_normal_modes_cartesian_shape(cls, v: SerializableNDArray):
        if v is not None:
            # Assume array has length of the number of normal modes
            n_normal_modes = len(v)
            return np.asarray(v).reshape(n_normal_modes, -1, 3)

    @field_validator("gradient")
    @classmethod
    def validate_gradient_shape(cls, v: SerializableNDArray):
        """Validate gradient is n x 3"""
        if v is not None:
            return np.asarray(v).reshape(-1, 3)

    @field_validator("hessian")
    @classmethod
    def validate_hessian_shape(cls, v: SerializableNDArray):
        """Validate hessian is square"""
        if v is not None:
            v = np.asarray(v)
            n = int(np.sqrt(v.size))
            return v.reshape((n, n))

    def return_result(self, calctype: CalcType) -> Union[float, SerializableNDArray]:
        """Return the primary result of the calculation."""
        return getattr(self, calctype.value)

    @model_validator(mode="after")
    def ensure_results(self) -> Self:
        """Ensure that at least one result is present."""
        if all(
            result is None
            for result in [
                self.energy,
                self.gradient,
                self.hessian,
            ]
        ):
            raise ValueError(
                "SinglePointResults requires either an energy, gradient, or hessian "
                "value."
            )
        return self


class OptimizationResults(QCIOModelBase):
    """Computed properties for an optimization.

    Attributes:
        energies: The energies for each step of the optimization.
        structures: The Structure objects for each step of the optimization.
        final_structure: The final, optimized Structure.
        trajectory: The SinglePointOutput objects for each step of the optimization.
    """

    trajectory: List[
        Union[
            ProgramOutput[ProgramInput, SinglePointResults],
            ProgramOutput[ProgramInput, NoResults],
        ]
    ] = []

    @property
    def final_structure(self) -> Optional[Structure]:
        """The final Structure in the optimization."""
        try:
            return self.trajectory[-1].input_data.structure
        except IndexError:  # Empty trajectory
            return None

    @property
    def final_molecule(self) -> Optional[Structure]:
        warnings.warn(
            ".final_molecule is being depreciated and will be removed in a future. "
            "Please use .final_structure instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        return self.final_structure

    @property
    def final_energy(self) -> Optional[Structure]:
        """The final energy in the optimization."""
        return self.energies[-1]

    @property
    def energies(self) -> np.ndarray:
        """The energies for each step of the optimization."""
        energies = np.empty(len(self.trajectory), dtype=float)

        for i, output in enumerate(self.trajectory):
            if not isinstance(output.results, NoResults):
                energies[i] = output.results.energy
            else:
                # If the calculation failed, set the energy to nan
                energies[i] = np.nan

        return energies

    @property
    def structures(self) -> List[Structure]:
        """The Structure objects for each step of the optimization."""
        return [output.input_data.structure for output in self.trajectory]

    @property
    def molecules(self) -> List[Structure]:
        """The Structure objects for each step of the optimization."""
        warnings.warn(
            ".molecules is being depreciated and will be removed in a future. "
            "Please use .structures instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        return self.structures

    def return_result(self, calctype: CalcType) -> Optional[Structure]:
        """Return the primary result of the calculation."""
        return self.final_structure

    def __repr_args__(self):
        """Custom repr to avoid printing the entire collection of objects."""
        return [
            ("final_structure", f"{self.final_structure}"),
            ("trajectory", "[...]"),
            ("energies", "[...]"),
            ("structures", "[...]"),
        ]

    def to_xyz(self) -> str:
        """Return the trajectory as an XYZ file."""
        return "".join(
            prog_output.input_data.structure.to_xyz() for prog_output in self.trajectory
        )

    def save(
        self,
        filepath: Union[Path, str],
        exclude_none=True,
        indent: int = 4,
        **kwargs,
    ) -> None:
        """Save an OptimizationOutput to a file.

        Args:
            filepath: The path to save the molecule to.
            exclude_none: If True, attributes with a value of None will not be written
                to the file.
            **kwargs: Additional keyword arguments to pass to the json serializer.

        Notes:
            If the filepath has a .xyz extension, the trajectory will be saved to an XYZ
            file.
        """
        filepath = Path(filepath)
        if filepath.suffix == ".xyz":
            filepath.write_text(self.to_xyz())
            return
        super().save(filepath, exclude_none, indent, **kwargs)


Results = Union[NoResults, SinglePointResults, OptimizationResults]
ResultsType = TypeVar("ResultsType", bound=Results)


class ProgramOutput(Files, Generic[InputType, ResultsType]):
    input_data: InputType
    provenance: Provenance
    success: Literal[True, False]
    results: ResultsType = NoResults()  # type: ignore
    stdout: Optional[str] = None
    traceback: Optional[str] = None

    def model_post_init(self, __context) -> None:
        """Parameterize the class (if not set explicitly)."""
        # Check if the current class is still generic, do not override if explicitly set
        if self.__class__ is ProgramOutput:
            input_type = type(self.input_data)
            # TODO: Make sure this is valid for results == None
            results_type = type(self.results)
            self.__class__ = ProgramOutput[input_type, results_type]  # type: ignore # noqa 501

    @model_validator(mode="after")
    def ensure_traceback_on_failure(self) -> Self:
        if self.success is False and self.traceback is None:
            raise ValueError("A traceback must be provided for failed calculations.")
        return self

    @model_validator(mode="after")
    def ensure_noresults_on_failure_for_single_point(self) -> Self:
        """Ensure that failed single point calculations have a NoResult object.

        Failed optimization or transition_state calculations may still have a partially
        completed trajectory as an OptimizationResults object.
        """
        if (
            type(self.input_data) is not FileInput
            and not self.success
            # NOTE: Must # type: ignore because the type check above is not respected
            # by mypy. Perhaps I'll want to reconsider how inputs inherit from FileInput
            # so I can use isinstance() checks??
            and self.input_data.calctype  # type: ignore
            in {
                CalcType.energy,
                CalcType.gradient,
                CalcType.hessian,
            }
        ):
            assert isinstance(self.results, NoResults), (
                "Unsuccessful energy, gradient and hessian calculations must set "
                ".results to NoResults."
            )

        return self

    @model_validator(mode="after")
    def ensure_results_on_success(self) -> Self:
        if self.success is True and type(self.input_data) is not FileInput:
            # Ensure results are provided for successful calculations
            assert not isinstance(
                self.results, NoResults
            ), "Results must be provided for successful, non FileInput calculations."

            if type(self.results) is SinglePointResults:
                # Ensure the primary calctype result is present
                calctype_val = self.input_data.calctype.value  # type: ignore
                assert (
                    getattr(self.results, calctype_val) is not None
                ), f"Missing the primary result: {calctype_val}."

        return self

    @property
    def pstdout(self) -> None:
        """Print the stdout text"""
        print(self.stdout)

    @property
    def ptraceback(self) -> None:
        """Print the traceback text"""
        print(self.traceback)

    def __repr_args__(self) -> List[Tuple[str, Any]]:
        """Exclude stdout and traceback from the repr and ensure success is first"""
        args = super().__repr_args__()

        # Replace stdout and traceback with "<...>"
        filtered_args = [
            (key, value if key not in {"stdout", "traceback"} else "<...>")
            for key, value in args
        ]
        # Ensure success is first
        success_arg = [(key, value) for key, value in filtered_args if key == "success"]
        other_args = [(key, value) for key, value in filtered_args if key != "success"]
        return success_arg + other_args

    @property
    def return_result(self) -> Union[float, SerializableNDArray, Optional[Structure]]:
        """Return the primary result of the calculation."""
        warnings.warn(
            ".return_result is being depreciated and will be removed in a future. "
            "Please access results directly at .results instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        # For mypy
        assert self.results is not None, "No results exist on this ProgramOutput"
        assert type(self.input_data) is not FileInput, "FileInputs have no results."
        return self.results.return_result(self.input_data.calctype)  # type: ignore


# For compatibility with old API
@deprecated_class("ProgramOutput[StructuredInputs, OptimizationResults]")
class OptimizationOutput(ProgramOutput[DualProgramInput, OptimizationResults]):
    success: Literal[True] = True
    traceback: Optional[str] = None


@deprecated_class("ProgramOutput[ProgramInput, SinglePointResults]")
class SinglePointOutput(ProgramOutput[ProgramInput, SinglePointResults]):
    success: Literal[True] = True
    traceback: Optional[str] = None


@deprecated_class("ProgramOutput[StructuredInputs, Results]")
class ProgramFailure(ProgramOutput[Inputs, Results]):
    success: Literal[False] = False


# Register the concrete classes for serialization
def _register_program_output_classes():
    """Required so that pickle can find the concrete classes for serialization."""
    for input_type, results_type in product(get_args(Inputs), get_args(Results)):
        _class = ProgramOutput[input_type, results_type]
        name = _class.__name__
        this_module = sys.modules[__name__]
        if name not in this_module.__dict__:
            # Directly declare the type to ensure it is registered
            setattr(this_module, name, _class)


# Call this function during module initialization
_register_program_output_classes()



import json
from typing import List, Union

import numpy as np
from pydantic import BaseModel

from .models import Structure  # FileOutput,

# Helper Structures
water = Structure(
    symbols=["O", "H", "H"],
    geometry=np.array([0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0, 1.0, 0.0]),
    charge=0,
    multiplicity=1,
    connectivity=[(0, 1, 1.0), (0, 2, 1.0)],
)


def json_dumps(obj: Union[BaseModel, List[BaseModel]]) -> str:
    """Serialization helper for lists of pydantic objects."""
    if isinstance(obj, list):
        return json.dumps([o.model_dump() for o in obj])
    return obj.model_dump_json()
